@@ -9,7 +9,7 @@ from imghdr import what
 from inspect import cleandoc
 from json import dumps, loads
 from logging import DEBUG, INFO, basicConfig, getLogger
-from os import environ
+from os import environ, mkdir
 from os.path import exists
 from pathlib import Path
 from sys import exit, stdout
@@ -25,9 +25,11 @@ class SimpleDatabase(dict):
     ease of hacking and accessibility and not for performance or efficiency.
     """
 
-    def __init__(self, filename, *args, **kwargs):
+    def __init__(self, filename, log, *args, **kwargs):
         """Initialise the object."""
         self.filename = Path(filename).absolute()
+        self.log = log
+
         self._loads()
         self.update(*args, **kwargs)
 
@@ -41,27 +43,27 @@ class SimpleDatabase(dict):
                 self.update(loads(handle.read()))
         except Exception as exception:
             message = f"Loading file storage failed: {exception}"
-            self.log.debug(message)
+            self.log.error(message, exc_info=exception)
             exit(1)
 
     def _dumps(self):
         """Save the databse to disk."""
         try:
             with open(self.filename, "w") as handle:
-                handle.write(dumps(self))
+                handle.write(dumps(self, indent=2, sort_keys=True))
         except Exception as exception:
             message = f"Saving file storage failed: {exception}"
-            self.log.debug(message)
+            self.log.error(message, exc_info=exception)
             exit(1)
 
     def __setitem__(self, key, val):
         """Write data to the database."""
-        dict.__setitem__(self, key, val)
+        super().__setitem__(key, val)
         self._dumps()
 
     def __delitem__(self, key):
         """Remove data from the database."""
-        dict.__delitem__(key)
+        super().__delitem__(key)
         self._dumps()
 
     def update(self, *args, **kwargs):
@@ -74,10 +76,11 @@ class SimpleDatabase(dict):
 class SimpleMessage:
     """A simple message interface."""
 
-    def __init__(self, message, bot):
+    def __init__(self, message, nick, log):
         """Initialise the object."""
         self.message = message
-        self.bot = bot
+        self._nick = nick
+        self.log = log
 
     @property
     def text(self):
@@ -103,12 +106,13 @@ class SimpleMessage:
         body = self.message["body"]
 
         try:
-            match = fr"^{self.bot.nick}.?(\s)"
+            match = fr"^{self._nick}.?(\s)"
             split = re.split(match, body)
             filtered = list(filter(None, split))
             return filtered[-1].strip()
         except Exception as exception:
-            self.bot.log.error(f"Couldn't parse {body}: {exception}")
+            message = f"Couldn't parse {body}: {exception}"
+            self.log.error(message, exc_info=exception)
             return None
 
     @property
@@ -212,9 +216,9 @@ class Config:
         return self.section.get("storage", None)
 
     @property
-    def storage_file(self):
-        """Path to the file based storage back-end."""
-        return self.section.get("storage_file", None)
+    def output(self):
+        """Path to the output directory."""
+        return self.section.get("output", None)
 
 
 class Bot(ClientXMPP):
@@ -236,6 +240,7 @@ class Bot(ClientXMPP):
         self.register_xmpp_event_handlers()
         self.register_xmpp_plugins()
         self.init_storage()
+        self.init_data()
         self.run()
 
     def parse_arguments(self):
@@ -321,10 +326,10 @@ class Bot(ClientXMPP):
             choices=("file", "redis"),
         )
         self.parser.add_argument(
-            "-stf",
-            "--storage-file",
-            dest="storage_file",
-            help="path to file based storage back-end",
+            "-o",
+            "--output",
+            dest="output",
+            help="path to output directory",
         )
 
         self.args = self.parser.parse_args()
@@ -380,15 +385,15 @@ class Bot(ClientXMPP):
         )
         rooms = input("Rooms: ")
 
-        if rooms:
-            rooms = [r.strip() for r in rooms.split(",")]
-
         inputs = {
             "account": account,
             "password": password,
             "nick": nick,
-            "rooms": rooms,
         }
+
+        if rooms:
+            inputs["rooms"] = rooms
+
         self.log.debug(f"Received {inputs} as input")
 
         config = ConfigParser()
@@ -457,11 +462,11 @@ class Bot(ClientXMPP):
             or environ.get("XBOT_STORAGE", None)
             or "file"
         )
-        storage_file = (
-            self.args.storage_file
-            or self.config.storage_file
-            or environ.get("XBOT_STORAGE_FILE", None)
-            or f"{nick}.json"
+        output = (
+            self.args.output
+            or self.config.output
+            or environ.get("XBOT_OUTPUT", None)
+            or "."
         )
 
         if not account:
@@ -491,7 +496,7 @@ class Bot(ClientXMPP):
             self.template = self.load_template(template)
 
         self.storage = storage
-        self.storage_file = Path(storage_file).absolute()
+        self.output = Path(output).absolute()
 
     def load_template(self, template):
         """Load template via Jinja."""
@@ -500,17 +505,22 @@ class Bot(ClientXMPP):
 
         try:
             from jinja2 import Environment, FileSystemLoader
-        except ModuleNotFoundError:
-            print("Missing required dependency: jinja2")
-            print("Have you tried `pip install xbotlib[web]`")
+        except ModuleNotFoundError as exception:
+            self.log.error(
+                "Missing required dependency jinja2, ",
+                "have you tried `pip install xbotlib[web]`",
+                exc_info=exception,
+            )
             exit(1)
 
         try:
             loader = FileSystemLoader(searchpath="./")
             env = Environment(loader=loader)
             return env.get_template(template)
-        except Exception:
-            self.log.info(f"Unable to load {template}")
+        except Exception as exception:
+            message = f"Unable to load {template}"
+            self.log.error(message, exc_info=exception)
+            exit(1)
 
     def register_xmpp_event_handlers(self):
         """Register functions against specific XMPP event handlers."""
@@ -521,12 +531,12 @@ class Bot(ClientXMPP):
         self.add_event_handler("message_error", self.error_message)
 
     def error_message(self, message):
-        message = SimpleMessage(message, self)
+        message = SimpleMessage(message, self.nick, self.log)
         self.log.error(f"Received error message: {message.text}")
 
     def direct_message(self, message):
         """Handle direct message events."""
-        message = SimpleMessage(message, self)
+        message = SimpleMessage(message, self.nick, self.log)
 
         if message.type not in self.DIRECT_MESSAGE_TYPES:
             return
@@ -542,7 +552,8 @@ class Bot(ClientXMPP):
         try:
             self.direct(message)
         except Exception as exception:
-            self.log.info(f"Bot.direct threw exception {exception}")
+            message = f"Bot.direct threw exception: {exception}"
+            self.log.error(message, exc_info=exception)
 
         if self.storage == "file":
             self.db._dumps()
@@ -556,12 +567,18 @@ class Bot(ClientXMPP):
 
     def publish_avatar(self):
         """Publish bot avatar."""
+        abspath = Path(self.avatar).absolute()
+
+        if not exists(abspath):
+            self.log.info(f"No avatar discovered (tried '{abspath})'")
+            return
+
         try:
-            abspath = Path(self.avatar).absolute()
             with open(abspath, "rb") as handle:
                 contents = handle.read()
-        except IOError:
-            self.log.info(f"No avatar discovered (tried '{abspath}')")
+        except Exception as exception:
+            message = f"Failed to load avatar: {exception}"
+            self.log.error(message, exc_info=exception)
             return
 
         id = self.plugin["xep_0084"].generate_id(contents)
@@ -583,6 +600,10 @@ class Bot(ClientXMPP):
             self.plugin["xep_0045"].join_muc(room, self.config.nick)
             self.log.info(f"Joining {room} automatically")
 
+        for room in self._data["invited"]:
+            self.plugin["xep_0045"].join_muc(room, self.config.nick)
+            self.log.info(f"Re-joining {room} (invited previously)")
+
     def group_invite(self, message):
         """Accept invites to group chats."""
         room = message["from"]
@@ -593,9 +614,13 @@ class Bot(ClientXMPP):
         self.plugin["xep_0045"].join_muc(room, self.config.nick)
         self.log.info(f"Joining {room} as invited")
 
+        if room not in self._data["invited"]:
+            self._data["invited"].append(str(room))
+            self._data._dumps()
+
     def group_message(self, message):
         """Handle group chat message events."""
-        message = SimpleMessage(message, self)
+        message = SimpleMessage(message, self.nick, self.log)
 
         if "@" in message.text:
             if self.meta(message, room=message.room):
@@ -619,7 +644,8 @@ class Bot(ClientXMPP):
         try:
             self.group(message)
         except Exception as exception:
-            self.log.info(f"Bot.group threw exception: {exception}")
+            message = f"Bot.group threw exception: {exception}"
+            self.log.error(message, exc_info=exception)
 
         if self.storage == "file":
             self.db._dumps()
@@ -641,18 +667,20 @@ class Bot(ClientXMPP):
                 self.register_plugin(plugin)
                 self.log.info(f"Loaded {plugin}")
         except Exception as exception:
-            message = f"Loading additional plugins failed {exception}"
-            self.log.info(message)
+            message = f"Loading additional plugins failed: {exception}"
+            self.log.error(message, exc_info=exception)
 
     def init_storage(self):
         """Initialise the storage back-end."""
+        file_storage_path = f"{self.output}/{self.nick}.json"
+
         if self.storage == "file":
             try:
-                self.db = SimpleDatabase(self.storage_file)
+                self.db = SimpleDatabase(file_storage_path, self.log)
                 self.log.info("Successfully loaded file storage")
             except Exception as exception:
-                message = f"Failed to load {self.storage_file}: {exception}"
-                self.log.info(message)
+                message = f"Failed to load {file_storage_path}: {exception}"
+                self.log.error(message, exc_info=exception)
                 exit(1)
         else:
             try:
@@ -662,13 +690,40 @@ class Bot(ClientXMPP):
                 return self.log.info("Successfully connected to Redis storage")
             except ValueError as exception:
                 message = f"Failed to connect to Redis storage: {exception}"
-                self.log.info(message)
+                self.log.error(message, exc_info=exception)
                 exit(1)
-            except ModuleNotFoundError:
-                print("missing required dependency using Redis")
-                print("Have you tried `pip install xbotlib[redis]`")
+            except ModuleNotFoundError as exception:
+                self.log.error(
+                    "Missing required dependency using Redis, ",
+                    "have you tried `pip install xbotlib[redis]`",
+                    exc_info=exception,
+                )
                 exit(1)
 
+    def init_data(self):
+        """Initialise internal data storage."""
+        try:
+            internal_dir = Path(".xbotlib").absolute()
+            if not exists(internal_dir):
+                mkdir(internal_dir)
+                self.log.info("Successfully initialised internal directory")
+        except Exception as exception:
+            message = f"Failed to create {internal_dir}: {exception}"
+            self.log.error(message, exc_info=exception)
+            exit(1)
+
+        try:
+            internal_storage_path = f"{internal_dir}/data.json"
+            self._data = SimpleDatabase(internal_storage_path, self.log)
+            self.log.info("Successfully loaded internal storage")
+        except Exception as exception:
+            message = f"Failed to load {internal_storage_path}: {exception}"
+            self.log.error(message, exc_info=exception)
+            exit(1)
+
+        if "invited" not in self._data:
+            self._data["invited"] = []
+
     def run(self):
         """Run the bot."""
         self.connect()
@@ -684,7 +739,7 @@ class Bot(ClientXMPP):
                     self.log.info("Finished running setup")
                 except Exception as exception:
                     message = f"Bot.setup failed: {exception}"
-                    self.log.info(message)
+                    self.log.error(message, exc_info=exception)
 
             self.process(forever=False)
         except (KeyboardInterrupt, RuntimeError):
@@ -694,9 +749,12 @@ class Bot(ClientXMPP):
         """Run the web server."""
         try:
             from aiohttp.web import Application, get, run_app
-        except ModuleNotFoundError:
-            print("Missing required dependency: aiohttp")
-            print("Have you tried `pip install xbotlib[web]`")
+        except ModuleNotFoundError as exception:
+            self.log.error(
+                "Missing required dependency aiohttp, ",
+                "have you tried `pip install xbotlib[web]`",
+                exc_info=exception,
+            )
             exit(1)
 
         self.web = Application()
@@ -719,9 +777,12 @@ class Bot(ClientXMPP):
         """Default placeholder text for HTML serving."""
         try:
             from aiohttp.web import Response
-        except ModuleNotFoundError:
-            print("Missing required dependency: aiohttp")
-            print("Have you tried `pip install xbotlib[web]`")
+        except ModuleNotFoundError as exception:
+            self.log.error(
+                "Missing required dependency aiohttp, ",
+                "have you tried `pip install xbotlib[web]`",
+                exc_info=exception,
+            )
             exit(1)
 
         return Response(text=f"{self.nick} is alive and well")
@@ -771,9 +832,12 @@ class Bot(ClientXMPP):
         """Send this response back with the web server."""
         try:
             from aiohttp.web import Response
-        except ModuleNotFoundError:
-            print("Missing required dependency: aiohttp")
-            print("Have you tried `pip install xbotlib[web]`")
+        except ModuleNotFoundError as exception:
+            self.log.error(
+                "Missing required dependency aiohttp, ",
+                "have you tried `pip install xbotlib[web]`",
+                exc_info=exception,
+            )
             exit(1)
 
         return Response(text=response, content_type=content_type)
